;    aperture by unlord / xylem
;          256b intro for
;       Function 2018 Party
;
; nasm aperture.asm -o aperture.com

struc stack
  .di: resw 1
  .si: resw 1
  .bp: resw 1
  .sp: resw 1
  .bx: resw 1
  .dx: resw 1
  .cx: resw 1
  .ax: resw 1
  .size:
endstruc

; Source texture is translated, rotated and then scaled.
;
; | zx  0  0 |   | c -s  0 |   | 1  0 -dx |   | x |   | u |
; |  0 zy  0 | * | s  c  0 | * | 0  1 -dy | * | y | = | v |
; |  0  0  1 |   | 0  0  1 |   | 0  0   1 |   | 1 |   | 1 |
;
; We want to invert this for roto zoomer.
;
; | 1  0  dx |   | c  s  0 |   | zx'  0  0 |   | u |   | x |
; | 0  1  dy | * |-s  c  0 | * |  0  zy' 0 | * | v | = | y |
; | 0  0   1 |   | 0  0  1 |   |  0   0  1 |   | 1 |   | 1 |
;
;         | 1  0  dx |   | c*zx'  s*zy'  0 |   | u |   | x |
;         | 0  1  dy | * |-s*zx'  c*zy'  0 | * | v | = | y |
;         | 0  0   1 |   |   0      0    1 |   | 1 |   | 1 |

  org 0x100

%define WIDTH 256
%define HEIGHT 204
%define STRIDE 320
%define TEXTURE 0x5000
%define SCREEN 0xa000

; Set initial register values zero'd when run in TD.EXE (costs 5 bytes)
;%define TD_SAFE

; Generate an XOR texture (costs 13 bytes)
%define GEN_TEXTURE

; Generate a texture of random noise (costs 15 bytes)
;%define RAND_TEXTURE

; Display the rotation center for debugging (costs 14 bytes)
;%define SHOW_CENTER

; Advance time faster for debugging in DOSBOX (costs 0 bytes)
;%define TIME_FAST

; Carry rounding to next pixel when blurring (costs 9 bytes)
;%define DIRTY_BLUR

; Cheaper version of DIRTY_BLUR (costs 6 bytes)
%define DIRTY_BYTE

; Waits for vertical retrace (costs 13 bytes)
%define WAIT_VSYNC

; Waits for two vertical retrace per frame (costs 3 bytes)
;%define DOUBLE_VSYNC

; This hides the blur on the left and right edges (costs 11 bytes)
;%define CLEAN_SIDES

; Display the palette at the top of the scren (costs 23 bytes)
;%define SHOW_PALETTE

; Set a grayscale palette (costs 15 bytes)
%define SET_PALETTE

; This adds a phase offset when generating a pair of sin values (costs 3 bytes)
%define USE_PHASE

; Fade the palette back and forth based on a sin value (costs 7 bytes)
%define FADE_PALETTE

; Draw + shaped dots before blurring (costs 29 bytes)
%define DRAW_DOTS

; Use square dots (saves 4 bytes)
;%define SQUARE_DOTS

; Loop the intro forever (saves 10 bytes)
%define LOOP_FOREVER

segment .text
  nop
  mov cl, HEIGHT
  xor di, di
  mov bx, si
  mov bp, sp

%ifdef TD_SAFE
  mov si, 0x100
  mov bx, si
%endif

  mov al, 0x13
  int 0x10

  pusha

%ifdef GEN_TEXTURE
  lea di, [bx + si]

%ifdef RAND_TEXTURE
  ; Fill texture with output from the "satisfactory" 16-bit pseudo-random
  ;  number generator (PRNG) in [1].
  ;
  ; [1] M. Yang, D. Robinson "Understanding and Learning Statistics by Computer"
  ;      World Scientific Publishing Company, Singapore (1986) page 6.
  mov ch, 0x7e
  mov bx, 0x6255
@rand:
  mul bx
  add ax, 0x3619
  stosw
  loop @rand

%else

  ; Fill texture with simple XOR pattern
  mov ch, 0xfc
@texture:
  mov ax, di
  xor al, ah
%ifdef DIRTY_BLUR OR DIRTY_BYTE
  and al, 0x3f
%endif
  stosb
  loop @texture
%endif
%endif

%ifdef SET_PALETTE
  ; CX = 0
@pal:
  mov ax, cx
  mov dx, 0x3C8
  out dx, al
  inc dx
  shr al, 2
  out dx, al
  shr al, 1
  out dx, al
  add al, 0x8
  out dx, al
  loop @pal
%endif

  popa

  ; AX = 0x13
  ; BX = 0x100
  ; DI = 0
  ; SI = 0x100
  ; BP = SP

@main:

  inc ax

  ;---------------------
  ; Stage 0: Roto A->B
  ;  DS = CS
  ;  ES = TEXTURE
  ;---------------------

_104:
  push TEXTURE
  pop es

  pusha

  ; Compute rotation center
  ;  st = t
  call gen_pair
  ; st = t, y0, x0

  ; Compute rotozoom constants (assume zoom = 1.0)
  ;  cz = cos(t)*zoom
  ;  sz = sin(t)*zoom
  fsincos
  ; st = c, s, y0, x0

  ; AX = row
  ; CX = col
  ; BX = rotation center
@row:
  mov ax, si
@col:
  pusha

  ; Compute the pixel position
  ;  x = sx - WIDTH/2
  ;  y = sy - HEIGHT/2
  add ax, -(WIDTH/2)
  add cx, -(HEIGHT/2)

  ; Compute the texture offset
  ;  u =  cz*(x + x0) + sz*(y + y0)
  ;  v = -sz*(x + x0) + cz*(y + y0)

  pusha

  push -1
  pop di
@again:
  ; st = cz, sz, y0, x0
  fild word [bp - 3*stack.size + stack.ax]
  ; st = x, cz, sz, y0, x0
  fadd st4
  ; st = x + x0, cz, sz, y0, x0
  fmul st1
  ; st = (x + x0)*cz, cz, sz, y0, x0
  fild word [bp - 3*stack.size + stack.cx]
  ; st = y, (x + x0)*cz, cz, sz, y0, x0
  fadd st4
  ; st = y + y0, (x + x0)*cz, cz, sz, y0, x0
  fmul st3
  ; st = (y + y0)*sz, (x + x0)*cz, cz, sz, y0, x0
  faddp
  ; st = (y + y0)*sz + (x + x0)*cz, cz, sz, y0, x0
  fistp word [bp - 3*stack.size + stack.si + 1 + di]
  ; st = cz, sz, y0, x0
@flip:
  fxch
  ; st = sz, cz, y0, x0
  fchs
  ; st = -sz, cz, y0, x0
  inc di
  jz  @again
  jpo @flip

  popa

  ; Add the rotation center
  add si, bx
  movsb

  popa
  inc di

  dec ax
  jnz @col
  loop @row

%ifdef SHOW_CENTER
  mov di, bx
  stosb
  add di, WIDTH - 2
  stosw
  stosb
  add di, WIDTH - 1
  stosb
%endif

  popa

  ;---------------------
  ; Stage 1: Draw on B
  ;  DS = CS
  ;  ES = TEXTURE
  ;---------------------

%ifdef DRAW_DOTS
  pusha

  mov cl, 75
@draw_dot:
  inc si
  call gen_pair
  pusha
  lodsb
  and al, 0x3f
  mov di, bx
%ifdef SQUARE_DOTS
  stosb
  stosb
  add di, 254
  stosb
  stosb
%else
  mov bx, 254
  stosb
  lea di, [di + bx]
  stosb
  stosb
  stosb
  lea di, [di + bx]
  stosb
%endif
  popa
  loop @draw_dot

  popa
%endif

  ;---------------------
  ; Stage 2: Blur B->A
  ;  DS = TEXTURE
  ;  ES = CS
  ;---------------------

  push es
  pop ds
  push cs
  pop es

  pusha

  lea di, [bx + si]
  add si, di
%ifdef DIRTY_BLUR
  mov ch, (HEIGHT - 4)/2
@blur:
  stosw
  lodsw
  add ax, dx
  add ax, [si + bx]
  add ax, [di]
  add ax, [si + 1]
  mov dx, 0x303
  and dx, ax
  sub ax, dx
  shr ax, 2
  loop @blur
%else
  xchg ch, cl
@blur:
  stosb
  lodsb
  add al, [si + bx]
  add al, [di]
  add al, [si + 1]
%ifdef DIRTY_BYTE
  add al, dl
  mov dl, 0x3
  and dl, al
%endif
  shr al, 2
  loop @blur
%endif
  popa

%ifdef CLEAN_SIDES
  pusha
  lea di, [bx + si]
@borders:
  stosw
  lea di, [di + bx - 4]
  stosw
  loop @borders
  popa
%endif

  ;---------------------
  ; Stage 3: Paint A
  ;  DS = TEXTURE
  ;  ES = SCREEN
  ;---------------------

  ; Draw the first line off screen
  push SCREEN - (STRIDE - (STRIDE - WIDTH)/2)/16
  pop es
  push cs
  pop ds

  pusha

%ifdef FADE_PALETTE
  call gen_pair
@draw:
  mov dx, 256
@draw_row:
  lodsb
  add al, bl
  stosb
  dec dx
  jnz @draw_row
  add di, STRIDE - WIDTH
  loop @draw

%else

@draw:
  mov dx, bx
@draw_row:
  movsb
  dec dx
  jnz @draw_row
  add di, STRIDE - WIDTH
  loop @draw

%endif

%ifdef SHOW_PALETTE
  pusha
  xor di, di
  mov cx, 32
@draw_palette:
  mov dx, 256
@draw_palette_row:
  mov al, 255
  sub al, dl
  stosb
  dec dx
  jnz @draw_palette_row
  add di, STRIDE-WIDTH
  loop @draw_palette
  popa
%endif

%ifdef WAIT_VSYNC
  mov dx, 0x3da
@virt:
  in al, dx
  test al, 8
  jnz @virt
@no_virt:
  in al, dx
  test al, 8
  jz @no_virt

%ifdef DOUBLE_VSYNC
  dec bx
  jp @virt
%endif
%endif

%ifdef LOOP_FOREVER
  popa
  jmp @main

%else

  mov ah, 1
  int 0x16
  popa
  jz @main

  mov ax, 0x3
  int 0x10

  ret
%endif

; Generate a pair of sin values from constants table (costs 23 bytes)
;
; in:
;  SI = pointer to constants table
;  DI = 0
;  BP = SP + 2*stack.size
;  st = t, ...
; out:
;  BL = 104*sin(w[si + 0]/32768*t + w[si + 2])
;  BH = 104*sin(w[si + 4]/32768*t + w[si + 6])
;  st = t, ...
gen_pair:
  pusha

  ; Compute the current time
  ;  DX = frame
  ;  SI = 256

  ; st = ?
  fninit
  ; st =
  fild word [bp - 2 - 2*stack.size + stack.ax]
  ; st = f
%ifdef TIME_FAST
  fidiv word [_104]
%else
  fidiv word [bp - stack.size + stack.si]
%endif

  dec di
@comp:
  fild word [si]
  lodsw
  ; st = a, t, ...
  fmul st1
  ; st = a*t, t, ...
%ifdef USE_PHASE
  fiadd word [si]
  lodsw
  ; st = a*t + b, t, ...
%endif
  ; st = v, t, ...
  fidiv word [_104 + 1]
  ; st = v/-32768, t, ...
  fsin
  ; st = s, t, ...
  fimul word [_104]
  ; st = 104*s, t
  fist word [bp - 2 - 2*stack.size + stack.bx + 1 + di]
  ; st = 104*s, t
  fxch
  ; st = t, 104*s
  inc di
  jz @comp
  popa
  add bx, (HEIGHT/2)*WIDTH + WIDTH/2
  ret
